#include <LeoArg.hh>

#include "CppDir.hh"
#include "xgetcwd.hh"
#include <iostream>
#include <string>
//#include <libintl.h>

#ifdef ENABLE_NLS
#  include <locale.h>
#    include <libintl.h>
#    define _(String) gettext (String)
#    ifdef gettext_noop
#        define N_(String) gettext_noop (String)
#    else
#        define N_(String) (String)
#    endif
#else
/* Stubs that do something close enough.  */
#    define textdomain(String) (String)
#    define gettext(String) (String)
#    define dgettext(Domain,Message) (Message)
#    define dcgettext(Domain,Message,Type) (Message)
#    define bindtextdomain(Domain,Directory) (Domain)
#    define _(String) (String)
#    define N_(String) (String)
#endif

using std::cout;
using std::endl;

void usage(); ///< print usage
void help();  ///< print help screen
void version(); ///< print the version number
void dirtree( std::string last_dir, std::string dir ); ///< recursive directory showing
void sort( CppDir::Directory::file_list& fl ); ///< sort files and directories
std::string print( CppDir::File::Type type ); ///< print a formated

static std::string progname;
static bool fullpath;     ///< if true the full path for each directory and file will be shown
static bool dir_fullpath; ///< if true the full path for each directory will be shown
static bool no_attribute;  ///< if true no directory attribute will be shown
static bool follow_links;  ///< follow links
static bool step_through;  ///< directory stepping
static bool view_links;    ///< mark the links
static bool absolute_path; ///< print absoloute paths
std::string current_path;  

int main( int argc, char** argv )
{
  setlocale(LC_ALL, "");
  bindtextdomain( PACKAGE, LOCALEDIR);
  textdomain(PACKAGE);

  LeoArg arg( argc, argv ); // initalize command line parser

  progname = arg.get_prog_name(); 

  arg.short_option_string( true );

  LeoArg::Option h( "help", "h" );
  if( arg.get_option( h ) )
    {
      help();
      return 0;
    }
    
 
  LeoArg::Option o( "no_attribute", "n" );
  no_attribute = arg.get_option( o );

  std::string dir;
  arg.get_last( dir );

  if( dir.empty() )
    dir = "."; // default dir is current dir
  
  o = LeoArg::Option( "fullpath", "f" );
  fullpath = arg.get_option( o );

  o = LeoArg::Option( "dir_fullpath", "d" );
  dir_fullpath = arg.get_option( o );

  o = LeoArg::Option( "follow_links", "l" );
  follow_links = arg.get_option( o );

  o = LeoArg::Option( "step_through", "s" );
  step_through = arg.get_option( o );

  o = LeoArg::Option( "view_links", "v" );
  view_links = arg.get_option( o );
		      

  o = LeoArg::Option( "absolute_path", "a" );
  if( absolute_path = arg.get_option( o ) )
    current_path = xgetcwd();

  dirtree( "", dir ); // start searching

  return 0;
}

void usage()
{
  cout << progname << ": " << _("Usage: ") << progname 
       << _(" [options] [directory]") << endl;
}

void help()
{
  usage();
  version();
  cout << _("  A tool for displaying directory trees") << endl;
  cout << _("  directory           the start-directory (default current directory)") << endl;
  cout << _("   Options:") << endl;
  cout << _("  -h --help             Print this screen") << endl;
  cout << _("  -n --no_attribute     Print no file attributes") << endl;
  cout << _("  -f --fullpath         Print the full path name for each directory and file") << endl;
  cout << _("  -d --dir_fullpath     Print the full path at each directory") << endl;
  cout << _("  -l --follow_links     Follow links") << endl;
  cout << _("  -s --step_through     Step through directories") << endl;
  cout << _("  -v --view_links       View links") << endl;
  cout << _("  -a --absolute_path    Print absolute paths") << endl;
}

void dirtree( std::string last_dir, std::string dir )
{
  std::string path;

  if( last_dir.empty() )
    path = dir;
  else
    path = CppDir::concat_dir(last_dir, dir );

  CppDir::Directory d( path ); // read directory

  if( !d )
    {
      cout << progname << _(": Error: invalid directory name: ") << dir << endl;
      return;
    }

  CppDir::Directory::file_list fl = d.get_files();

  sort( fl ); // sort files

  for( CppDir::Directory::file_list_it it = fl.begin();
       it != fl.end(); it++ )
    {
      if( it->get_name() == "." )
	continue;

      if( it->get_name() == ".." )
	continue;

      if( !no_attribute )
	{
	  cout << print( it->get_type() );
	  if( view_links && it->is_link())
	    cout << "L";

	  cout << " ";
	}
   
      if( fullpath || (it->get_type() == CppDir::File::DIR && dir_fullpath ))
	{
	  std::string apath = path;

	  if( absolute_path )
	    apath = CppDir::concat_dir( current_path, path );

	  cout << CppDir::concat_dir( apath, it->get_name()) << endl;
	}
      else if( !fullpath )
	{
	  for( int i = 0; i < last_dir.size() + 1; i++ )
	    cout << " "; // add spaces
      
	  cout << it->get_name() << endl;
	}


      bool dive = true;
      
      if( it->is_link() )
	if( follow_links )
	  dive = true;
	else
	  dive = false;
      
      if( (it->get_type() ==  CppDir::File::DIR) && dive ) 
	{
	  // dive into subdirectories
	  dirtree( path, it->get_name() );

	  if( step_through )
	    {
	      if( !no_attribute )
		cout << print( CppDir::File::DIR ) << " ";

	      if( fullpath || dir_fullpath )
		{
		  std::string apath = path;
		  
		  if( absolute_path )
		    apath = CppDir::concat_dir( current_path, path );

		  cout << CppDir::concat_dir( apath, ".." ) << endl;
		}
	      else if( !fullpath )
		{
		  for( int i = 0; i < last_dir.size() + 1; i++ )
		    cout << " "; // add spaces
		  
		  cout << ".." << endl;
		}

	    } // if

	}  // if
      
    } // for
}

/** resorts a file_list.
    this is not an alphabetic sorting. It simple puts the files at the front of the
    list and puts the directories at the end of the list */
void sort( CppDir::Directory::file_list& fl) 
{
  CppDir::Directory::file_list files, dirs;

  for( CppDir::Directory::file_list_it it = fl.begin();
       it != fl.end(); it++ )
    {
      if( it->get_type() == CppDir::File::DIR )
	dirs.push_back( *it );
      else
	files.push_back( *it );
    }

  fl.clear();

  for( CppDir::Directory::file_list_it it = files.begin();
       it != files.end(); it++ )
    fl.push_back( *it );

  for( CppDir::Directory::file_list_it it = dirs.begin();
       it != dirs.end(); it++ )
    fl.push_back( *it );
}

std::string print( CppDir::File::Type type )
{
  std::string ret;

  switch( type )
    {
    case CppDir::File::UNKNOWN: ret = "U"; break;
    case CppDir::File::FIFO   : ret = "F"; break;
    case CppDir::File::CHAR   : ret = "C"; break;
    case CppDir::File::DIR    : ret = "D"; break;
    case CppDir::File::BLOCK  : ret = "B"; break;
    case CppDir::File::REGULAR: ret = "R"; break;
    case CppDir::File::LINK   : ret = "L"; break;
    }

  return ret;
}

void version()
{
  cout << progname << _(": Version: ") << VERSION << endl;
}
